/* Copyright (C)  Walter L. Peacock 1984-1988.  All rights reserved */
/*  C B T R E E . C     8/02/84                     */

/* THIS PROGRAM BELONGS TO WALTER L. PEACOCK.  IT IS CONSIDERED A TRADE */
/* SECRET AND IS NOT TO BE DIVULGED OR USED BY PARTIES WHO HAVE NOT    */
/* RECEIVED WRITTEN AUTHORIZATION FROM THE OWNER.                   */
/*---------------------------------------------------------------------------- 
 * The driver module, cbtree.(), is the primary module used to access the 
 * B+tree.   This module provides a single function call interface 
 * to the CBTREE functions.  The cbtree() driver module determines which 
 * function to invoke automatically, based on the operation that is passed in 
 * btcommo btoptype member. 
 *
 * The functions are built on the lower-level B+tree maintenance functions which
 * are completely managed by the cbtree() driver module.  cbtree() returns a 
 * result code to indicate the status of the operation.   The result codes are 
 * defined in the cbtree.h file and were described in PART III of the manual.
 *
 * After the call to CBTREE, the record location is returned in btcommo btloc,
 * unless a block retrieval call (GETALL or GETSEQ) is requested, in which 
 * case the record locations will be returned in the btcommo array btrecnum.
 *---------------------------------------------------------------------------*/

#include "cbtree.h"

#if CI | DC
#else
#include <ctype.h>
#endif

long int locklist[20];
int lockptr;

int cbtree(fddta, fdidx, btcp)
int fddta, fdidx;
BTC *btcp;
{
   extern long lseek(), getroot(), poptop(), pushdown();
   extern char *strsave(), *calloc(), *malloc(), *strcpy(), *strnncpy();
   extern void free();
   extern int strlen(), btrgets(), wrap_hdr();
   BTIDXBLK *idxblk;
   long pos, node, rootptr;
   char *keyp;   /* a key pointer for general use */
   int lk, lockchk, hdrloc, retcode;
   unsigned sl, dupflg, match;
   register int i;

   lockptr = 0;
   i = sizeof(BTIDXBLK) + btcp->btcells * sizeof(BTIDX);
   if ((char *)(idxblk = (BTIDXBLK *)calloc(i, SZCHAR)) == NULL)
      ckerror(- CK_NOMEM, "40");
   btcp->btkey[btcp->btkeylen] = NUL;    /* make sure we have EOS */

   /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
   /* record locking is available via the function lockunlk().        */
   /* lockunlk() must be modified to function on your specific operating  */
   /* system and application.  It should return 1 if a successful lock/   */
   /* unlock occurred or 0 if the call was unsuccessful.              */
   /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

   if (btcp->btoptype == INSERT   ||  btcp->btoptype == ISRTKY)
      lockchk = 1;
   else if (btcp->btoptype == DELETE   ||  btcp->btoptype == DELTKY)
      lockchk = 2;
   else
      lockchk = 0;
   if(lockchk) {
      lk = 0;
      while (lockunlk(fdidx,(long)btcp->findroot, SZ_BT_INFO, LOCKR) != 1)
         if (++lk > HDRTRYS) {
            retcode = BTLOCK;
            goto cb_exit;
         }
      locklist[lockptr++] = (long) btcp->findroot;  /* save root node */
      lk = 0;
      while (lockunlk(fdidx, 0L, SZ_BT_INFO, LOCKR) != 1)
         if (++lk > HDRTRYS){
            i = lockunlk(fdidx, (long)btcp->findroot, SZ_BT_INFO, UNLOCKR);
            retcode = BTLOCK;
            goto cb_exit;
         }
      locklist[lockptr++] = 0L;    /* pointer to header record          */
   }

   /* get ptr to root node */
   rootptr = btcp->btroot = getroot(fdidx, btcp->findroot);

   if (pushdown(rootptr) < 1L) /* save root node on stack of longs */
      ckerror(- CK_PUSH, "41");
   if (lockchk){
      lk = 0;
      pos = (long) ( ((long) rootptr - 1L) * (long) btcp->btidxlen);
      while (lockunlk(fdidx, pos, btcp->btidxlen, LOCKR) != 1)
         if (++lk > HDRTRYS){
            retcode = BTLOCK;
            goto cb_exit;
         }
      locklist[lockptr++] = pos;
   }

   getidxr(fdidx, rootptr, idxblk, btcp->btidxlen, btcp->btkeylen);

   if (btcp->btoptype == GETLAST){   /* build the termination key */
      strcpy(btcp->btkey, "\177");
   }
   else
      if (btcp->btoptype == GETFRST){
         btcp->btkey[0] = NUL;
	  }

   /* if this is the first time thru a multi-pass operation, reset btloc */
   if (btcp->btsvkey == NULL &&
      (btcp->btoptype == GETALL || btcp->btoptype == GETKEYS))
      btcp->btloc = 0L;   /* caller should NOT modify this ! */

   /* select the proper key to use for a multi-pass operation */
   keyp = (btcp->btsvkey != NULL) ? btcp->btsvkey : btcp->btkey;
   sl = strlen(btcp->btkey);    /* the length of the callers key */

   /* traverse the index nodes until you find the desired data node */
   while (!idxblk->blktype) {   /* until data block type = 1 located */
      /* get offset into btpage for key */
      i = binarysr(idxblk, keyp, strlen(keyp));

      if(lockchk){
         lk = 0;
         pos = (idxblk->btpage[i].btptr - 1L) * btcp->btidxlen;
         while (lockunlk(fdidx, pos, btcp->btidxlen, LOCKR) != 1)
            if (++lk > HDRTRYS){
               retcode = BTLOCK;
               goto cb_exit;
            }
         locklist[lockptr++] = pos;
      }

      if (pushdown(idxblk->btpage[i].btptr) < 1L)    /* save current rec # */
         ckerror(- CK_PUSH, "42");

      getidxr(fdidx, idxblk->btpage[i].btptr,
              idxblk, btcp->btidxlen, btcp->btkeylen);
   }

   node = poptop();       /* get the rec #: if lockchk then node is locked */
   /* get match/insert position within hdr rec for given key   */
   hdrloc = binarydt(fdidx, idxblk, btcp, sl, &match, &dupflg);

   switch (btcp->btoptype) {
      case GETREC:
      case GETNXT:
      case GETPRV:
      case GETPAR:
      case GETALL:
      case GETSEQ:
      case GETKEYS:
         retcode = (match) ?
           btrgets(fdidx, idxblk, hdrloc, sl, &dupflg, btcp) : BTNOTFND;
         break;
      case GETGT:
      case GETGE:
      case GETLT:
      case GETLE:
         retcode = btrgets(fdidx, idxblk, hdrloc, sl, &dupflg, btcp);
         break;
      case GETLAST:
         if ((retcode = hdrloc =
                 wrap_hdr(hdrloc-1, idxblk, fdidx, btcp)) == BTNOTFND) {
            *btcp->btkey = NUL;
            goto cb_exit;
         }
         /* fall thru to GETFRST */
      case GETFRST:
         if ((btcp->btloc = idxblk->btpage[hdrloc].btptr) == 0L)
            retcode = BTNOTFND;
         else {
            strcpy(btcp->btkey, idxblk->btpage[hdrloc].skeynme);
            match = 1;
            retcode = BTCALLOK;
         }
         break;
      case NEWLOC:
         retcode = (match) ?
            btrisrt(fddta, fdidx, idxblk, hdrloc, node, rootptr, btcp) :
            BTNOTFND;
         break;
      case INSERT:
      case ISRTKY:
         retcode =
             btrisrt(fddta, fdidx, idxblk, hdrloc, node, rootptr, btcp);
         break;
      case DELETE:
      case DELTKY:
         retcode = (match) ?
           btrdelet(fddta, fdidx, idxblk, hdrloc, node,btcp) : BTNOTFND;
         break;
      default:
         puts("\nInvalid opcode used. Check btoptype. Press <ENTER>");
         retcode = BTCOMMO;
         break;
   }

cb_exit:   /* de-allocate everything and go home */
   freekeys(idxblk);
   FREE(idxblk);
   relstack();

   if(lockchk){
      while(--lockptr >= 0)
         lockunlk(fdidx, locklist[lockptr], lockptr < 2 ? SZ_BT_INFO : btcp->btidxlen, UNLOCKR);
   }

   return(retcode);
}

/* THIS PROGRAM BELONGS TO WALTER L. PEACOCK.  IT IS CONSIDERED A TRADE */
/* SECRET AND IS NOT TO BE DIVULGED OR USED BY PARTIES WHO HAVE NOT    */
/* RECEIVED WRITTEN AUTHORIZATION FROM THE OWNER.          */
